#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <cstdlib>
#include <sys/un.h>
#include <sys/socket.h>
#include <stdio.h>
#include <json-c/arraylist.h>

#include "str-vec.h"

#include "json-util.h"

JSON_OBJECT_T json_util::from_string(const std::string& str) {
	return json_object_new_string(str.c_str());
}

JSON_OBJECT_T json_util::from_int(int i) {
	return json_object_new_int(i);
}

JSON_OBJECT_T json_util::from_float(float flt) {
	return json_object_new_double(flt);
}

JSON_OBJECT_T json_util::from_boolean(bool b) {
	return json_object_new_boolean(b);
}

JSON_OBJECT_T json_util::empty() {
	return json_object_new_object();
}

template<typename T>
JSON_OBJECT_T json_object_from_list(const std::list<T>& list, JSON_OBJECT_T (*convert)(T)) {
	JSON_OBJECT_T jobj = json_object_new_array();
	for (
		typename std::list<T>::const_iterator it = list.begin();
		it != list.end();
		++it
	){
		json_object_array_add(jobj, convert(*it));
	}
	return jobj;
}

JSON_OBJECT_T json_util::from_string_list(const std::list<std::string>& list) {
	return json_object_from_list(list, json_util::from_value_string);
}

JSON_OBJECT_T json_util::from_string_vector(const string_vector& vect) {
	// TODO: Make sure this function does what it's supposed to do
	std::list<std::string> list;
	for (int i = 0; i < vect.numel(); ++i) {
		list.push_back(vect[i]);
	}

	return json_object_from_list(list, json_util::from_value_string);
}

JSON_OBJECT_T json_util::from_int_list(const std::list<int>& list) {
	return json_object_from_list(list, json_util::from_int);
}

JSON_OBJECT_T json_util::from_float_list(const std::list<float>& list) {
	return json_object_from_list(list, json_util::from_float);
}

JSON_OBJECT_T json_util::from_workspace_list(const std::list<workspace_element>& list) {
	return json_object_from_list(list, json_util::from_workspace_element);
}

JSON_OBJECT_T json_util::from_filter_list(const octave_link::filter_list& list) {
	return json_object_from_list(list, json_util::from_pair);
}

JSON_OBJECT_T json_util::from_value_string(const std::string str) {
	return json_util::from_string(str);
}

JSON_OBJECT_T json_util::from_workspace_element(workspace_element element) {
	JSON_MAP_T m;
	m["scope"] = json_util::from_int(element.scope());
	m["symbol"] = json_util::from_string(element.symbol());
	m["class_name"] = json_util::from_string(element.class_name());
	m["dimension"] = json_util::from_string(element.dimension());
	m["value"] = json_util::from_string(element.value());
	m["complex_flag"] = json_util::from_boolean(element.complex_flag());
	return json_util::from_map(m);
}

JSON_OBJECT_T json_util::from_pair(std::pair<std::string, std::string> pair) {
	JSON_OBJECT_T jobj = json_object_new_array();
	json_object_array_add(jobj, json_object_new_string(pair.first.c_str()));
	json_object_array_add(jobj, json_object_new_string(pair.second.c_str()));
	return jobj;
}

JSON_OBJECT_T json_util::from_map(JSON_MAP_T m) {
	JSON_OBJECT_T jobj = json_object_new_object();
	for(
		std::map<std::string, JSON_OBJECT_T>::iterator it = m.begin();
		it != m.end();
		++it
	){
		json_object_object_add(jobj, it->first.c_str(), it->second);
	}
	return jobj;
}

std::string json_util::to_message(const std::string& name, JSON_OBJECT_T jobj) {
  JSON_OBJECT_T jmsg = json_object_new_array();
  json_object_array_add(jmsg, json_util::from_string(name));
  json_object_array_add(jmsg, jobj);
  std::string str (json_object_to_json_string(jmsg));
  return str;
}

std::string json_util::to_string(JSON_OBJECT_T jobj) {
  return std::string(json_object_get_string(jobj));
}

template<typename T>
std::list<T> json_object_to_list(JSON_OBJECT_T jobj, T (*convert)(JSON_OBJECT_T)) {
	std::list<T> ret;

	struct array_list* arr = json_object_get_array(jobj);
	if (arr == NULL)
		return ret;

	for (int i = 0; i < array_list_length(arr); ++i) {
		JSON_OBJECT_T jsub = static_cast<JSON_OBJECT_T> (array_list_get_idx(arr, i));
		ret.push_back(convert(jsub));
	}
	return ret;
}

std::pair<std::list<int>, int> json_util::to_int_list_int_pair(JSON_OBJECT_T jobj) {
	std::pair<std::list<int>, int> ret;

	struct array_list* arr = json_object_get_array(jobj);
	if (arr == NULL)
		return ret;

	JSON_OBJECT_T first = static_cast<JSON_OBJECT_T> (array_list_get_idx(arr, 0));
	JSON_OBJECT_T second = static_cast<JSON_OBJECT_T> (array_list_get_idx(arr, 1));

	ret.first = json_object_to_list<int>(first, json_util::to_int);
	ret.second = json_object_get_int(second);

	return ret;
}

std::list<std::string> json_util::to_string_list(JSON_OBJECT_T jobj) {
	return json_object_to_list<std::string>(jobj, json_util::to_string);
}

int json_util::to_int(JSON_OBJECT_T jobj) {
	return json_object_get_int(jobj);
}

bool json_util::to_boolean(JSON_OBJECT_T jobj) {
	return json_object_get_boolean(jobj);
}

void json_util::read_stream(int sockfd, void (*cb)(std::string, JSON_OBJECT_T, void*), void* arg) {

	// Make some local variables
	int BUF_LEN = 24;
	char* buf = new char[BUF_LEN];  // buffer for socket read
	int buf_len;  // length of new bytes in the buffer
	int buf_offset;  // offset of the JSON parser in the buffer
	JSON_OBJECT_T jobj;  // pointer to parsed JSON object
	json_tokener* tok = json_tokener_new();  // JSON tokenizer instance
	enum json_tokener_error jerr;  // status of JSON tokenizer

	// Start the blocking I/O loop
	while( (buf_len=recv(sockfd, buf, BUF_LEN, 0)) > 0) {
		buf_offset = 0;
		while(buf_offset < buf_len){
			jobj = json_tokener_parse_ex(tok, buf + buf_offset, buf_len - buf_offset);
			jerr = json_tokener_get_error(tok);
			buf_offset += tok->char_offset;

			// Do we need more material in order to make JSON?
			if (jerr == json_tokener_continue) {
				continue;
			}

			// Make a new tokenizer
			json_tokener_free(tok);
			tok = json_tokener_new();

			// Did we encounter a malformed JSON object?
			if (jerr != json_tokener_success) {
				fprintf(stderr,
					"JSON parse error: %s: '%.*s'\n",
					json_tokener_error_desc(jerr),
					1,
					buf + buf_offset);
				fflush(stderr);
				break;
			}

			// Our object is ready
			process_message(jobj, cb, arg);
		}
	}

	json_tokener_free(tok);
	free(buf);
}

void json_util::process_message(JSON_OBJECT_T jobj, void (*cb)(std::string, JSON_OBJECT_T, void*), void* arg) {
  if (!json_object_is_type(jobj, json_type_array))
    return;
  if (json_object_array_length(jobj) != 2)
    return;

  cb(
  	json_util::to_string(json_object_array_get_idx(jobj, 0)),
  	json_object_array_get_idx(jobj, 1),
  	arg
  );
}
