/*

Copyright (C) 2015-2016 Shane Carr

This file is part of Octave.

Octave is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 3 of the License, or (at your
option) any later version.

Octave is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Octave; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream>
#include "octave-json-link.h"
#include "workspace-element.h"
#include "json-main.h"

// MAP_SET(m, foo, string)
//   => m["foo"] = json_util::from_string(foo);
#define MAP_SET(M, FIELD, TYPE){ \
	m[#FIELD] = json_util::from_##TYPE (FIELD); \
}

octave_json_link::octave_json_link(json_main* __json_main)
	: octave_link (),
		_json_main (__json_main)
{
	_request_input_enabled = true;
}

octave_json_link::~octave_json_link(void) { }

std::string octave_json_link::do_request_input(const std::string& prompt) {
	// Triggered whenever the console prompts for user input
	_publish_event("request-input", json_util::from_string(prompt));

	return request_input_queue.dequeue();
}

bool octave_json_link::do_confirm_shutdown(void) {
	// Triggered when the kernel tries to exit
	_publish_event("confirm-shutdown", json_util::empty());

	return confirm_shutdown_queue.dequeue();
}

bool octave_json_link::do_exit(int status) {
	JSON_MAP_T m;
	MAP_SET(m, status, int);
	_publish_event("exit", json_util::from_map(m));

	// It is our responsibility in octave_link to call exit. If we don't, then
	// the kernel waits for 24 hours expecting us to do something.
	::exit(status);

	return true;
}

bool octave_json_link::do_copy_image_to_clipboard(const std::string& file) {
	// This endpoint might be unused?  (References appear only in libgui)
	JSON_MAP_T m;
	MAP_SET(m, file, string);
	_publish_event("copy-image-to-clipboard", json_util::from_map(m));

	return true;
}

bool octave_json_link::do_edit_file(const std::string& file) {
	// Triggered in "edit" for existing files
	JSON_MAP_T m;
	MAP_SET(m, file, string);
	_publish_event("edit-file", json_util::from_map(m));

	return true;
}

bool octave_json_link::do_prompt_new_edit_file(const std::string& file) {
	// Triggered in "edit" for new files
	JSON_MAP_T m;
	MAP_SET(m, file, string);
	_publish_event("prompt-new-edit-file", json_util::from_map(m));

	return prompt_new_edit_file_queue.dequeue();
}

int octave_json_link::do_message_dialog(const std::string& dlg, const std::string& msg, const std::string& title) {
	// Triggered in "msgbox", "helpdlg", and "errordlg", among others
	JSON_MAP_T m;
	MAP_SET(m, dlg, string); // i.e., m["dlg"] = json_util::from_string(dlg);
	MAP_SET(m, msg, string);
	MAP_SET(m, title, string);
	_publish_event("message-dialog", json_util::from_map(m));

	return message_dialog_queue.dequeue();
}

std::string octave_json_link::do_question_dialog(const std::string& msg, const std::string& title, const std::string& btn1, const std::string& btn2, const std::string& btn3, const std::string& btndef) {
	// Triggered in "questdlg"
	JSON_MAP_T m;
	MAP_SET(m, msg, string);
	MAP_SET(m, title, string);
	MAP_SET(m, btn1, string);
	MAP_SET(m, btn2, string);
	MAP_SET(m, btn3, string);
	MAP_SET(m, btndef, string);
	_publish_event("question-dialog", json_util::from_map(m));

	return question_dialog_queue.dequeue();
}

std::pair<std::list<int>, int> octave_json_link::do_list_dialog(const std::list<std::string>& list, const std::string& mode, int width, int height, const std::list<int>& initial_value, const std::string& name, const std::list<std::string>& prompt, const std::string& ok_string, const std::string& cancel_string) {
	// Triggered in "listdlg"
	JSON_MAP_T m;
	MAP_SET(m, list, string_list);
	MAP_SET(m, mode, string);
	MAP_SET(m, width, int);
	MAP_SET(m, height, int);
	MAP_SET(m, initial_value, int_list);
	MAP_SET(m, name, string);
	MAP_SET(m, prompt, string_list);
	MAP_SET(m, ok_string, string);
	MAP_SET(m, cancel_string, string);
	_publish_event("list-dialog", json_util::from_map(m));

	return list_dialog_queue.dequeue();
}

std::list<std::string> octave_json_link::do_input_dialog(const std::list<std::string>& prompt, const std::string& title, const std::list<float>& nr, const std::list<float>& nc, const std::list<std::string>& defaults) {
	// Triggered in "inputdlg"
	JSON_MAP_T m;
	MAP_SET(m, prompt, string_list);
	MAP_SET(m, title, string);
	MAP_SET(m, nr, float_list);
	MAP_SET(m, nc, float_list);
	MAP_SET(m, defaults, string_list);
	_publish_event("input-dialog", json_util::from_map(m));

	return input_dialog_queue.dequeue();
}

std::list<std::string> octave_json_link::do_file_dialog(const filter_list& filter, const std::string& title, const std::string &filename, const std::string &pathname, const std::string& multimode) {
	// Triggered in "uiputfile", "uigetfile", and "uigetdir"
	JSON_MAP_T m;
	MAP_SET(m, filter, filter_list);
	MAP_SET(m, title, string);
	MAP_SET(m, filename, string);
	MAP_SET(m, pathname, string);
	MAP_SET(m, multimode, string);
	_publish_event("file-dialog", json_util::from_map(m));
	
	return file_dialog_queue.dequeue();
}

int octave_json_link::do_debug_cd_or_addpath_error(const std::string& file, const std::string& dir, bool addpath_option) {
	// This endpoint might be unused?  (No references)
	JSON_MAP_T m;
	MAP_SET(m, file, string);
	MAP_SET(m, dir, string);
	MAP_SET(m, addpath_option, boolean);
	_publish_event("debug-cd-or-addpath-error", json_util::from_map(m));

	return debug_cd_or_addpath_error_queue.dequeue();
}

void octave_json_link::do_change_directory(const std::string& dir) {
	// This endpoint might be unused?  (References appear only in libgui)
	JSON_MAP_T m;
	MAP_SET(m, dir, string);
	_publish_event("change-directory", json_util::from_map(m));
}

void octave_json_link::do_execute_command_in_terminal(const std::string& command) {
	// This endpoint might be unused?  (References appear only in libgui)
	JSON_MAP_T m;
	MAP_SET(m, command, string);
	_publish_event("execute-command-in-terminal", json_util::from_map(m));
}

void octave_json_link::do_set_workspace(bool top_level, bool debug, const std::list<workspace_element>& ws) {
	// Triggered on every new line entry
	JSON_MAP_T m;
	MAP_SET(m, top_level, boolean);
	MAP_SET(m, debug, boolean);
	MAP_SET(m, ws, workspace_list);
	_publish_event("set-workspace", json_util::from_map(m));
}

void octave_json_link::do_clear_workspace(void) {
	// Triggered on "clear" command (but not "clear all" or "clear foo")
	_publish_event("clear-workspace", json_util::empty());
}

void octave_json_link::do_set_history(const string_vector& hist) {
	// Called at startup, possibly more?
	JSON_MAP_T m;
	MAP_SET(m, hist, string_vector);
	_publish_event("set-history", json_util::from_map(m));
}

void octave_json_link::do_append_history(const std::string& hist_entry) {
	// Appears to be tied to readline, if available
	JSON_MAP_T m;
	MAP_SET(m, hist_entry, string);
	_publish_event("append-history", json_util::from_map(m));
}

void octave_json_link::do_clear_history(void) {
	// Appears to be tied to readline, if available
	_publish_event("clear-history", json_util::empty());
}

void octave_json_link::do_pre_input_event(void) {
	// noop
}

void octave_json_link::do_post_input_event(void) {
	// noop
}

void octave_json_link::do_enter_debugger_event(const std::string& file, int line) {
	JSON_MAP_T m;
	MAP_SET(m, file, string);
	MAP_SET(m, line, int);
	_publish_event("enter-debugger-event", json_util::from_map(m));
}

void octave_json_link::do_execute_in_debugger_event(const std::string& file, int line) {
	JSON_MAP_T m;
	MAP_SET(m, file, string);
	MAP_SET(m, line, int);
	_publish_event("execute-in-debugger-event", json_util::from_map(m));
}

void octave_json_link::do_exit_debugger_event(void) {
	_publish_event("exit-debugger-event", json_util::empty());
}

void octave_json_link::do_update_breakpoint(bool insert, const std::string& file, int line) {
	JSON_MAP_T m;
	MAP_SET(m, insert, boolean);
	MAP_SET(m, file, string);
	MAP_SET(m, line, int);
	_publish_event("update-breakpoint", json_util::from_map(m));
}

void octave_json_link::do_set_default_prompts(std::string& ps1, std::string& ps2, std::string& ps4) {
	// Triggered upon interpreter startup
	JSON_MAP_T m;
	MAP_SET(m, ps1, string);
	MAP_SET(m, ps2, string);
	MAP_SET(m, ps4, string);
	_publish_event("set-default-prompts", json_util::from_map(m));
}

void octave_json_link::do_show_preferences(void) {
	// Triggered on "preferences" command
	_publish_event("show-preferences", json_util::empty());
}

void octave_json_link::do_show_doc(const std::string& file) {
	// Triggered on "doc" command
	_publish_event("show-doc", json_util::from_string(file));
}

void octave_json_link::do_show_static_plot(const std::string& term, const std::string& content) {
	// Triggered on all plot commands with setenv("GNUTERM","svg")
	JSON_MAP_T m;
	MAP_SET(m, term, string);
	MAP_SET(m, content, string);
	_publish_event("show-static-plot", json_util::from_map(m));
}

void octave_json_link::receive_message (const std::string& name, JSON_OBJECT_T jobj) {
	if (name == "cmd" || name == "request-input-answer") {
		std::string answer = json_util::to_string(jobj);
		request_input_queue.enqueue(answer);
	}
	else if (name == "confirm-shutdown-answer"){
		bool answer = json_util::to_boolean(jobj);
		confirm_shutdown_queue.enqueue(answer);
	}
	else if (name == "prompt-new-edit-file-answer"){
		bool answer = json_util::to_boolean(jobj);
		prompt_new_edit_file_queue.enqueue(answer);
	}
	else if (name == "message-dialog-answer"){
		int answer = json_util::to_int(jobj);
		message_dialog_queue.enqueue(answer);
	}
	else if (name == "question-dialog-answer") {
		std::string answer = json_util::to_string(jobj);
		question_dialog_queue.enqueue(answer);
	}
	else if (name == "list-dialog-answer") {
		std::pair<std::list<int>, int> answer = json_util::to_int_list_int_pair(jobj);
		list_dialog_queue.enqueue(answer);
	}
	else if (name == "input-dialog-answer") {
		std::list<std::string> answer = json_util::to_string_list(jobj);
		input_dialog_queue.enqueue(answer);
	}
	else if (name == "file-dialog-answer") {
		std::list<std::string> answer = json_util::to_string_list(jobj);
		file_dialog_queue.enqueue(answer);
	}
	else if (name == "debug-cd-or-addpath-error-answer") {
		int answer = json_util::to_int(jobj);
		debug_cd_or_addpath_error_queue.enqueue(answer);
	}
	else {
		std::cerr << "warning: received unknown message: " << name << std::endl;
	}
}

void octave_json_link::_publish_event(const std::string& name, JSON_OBJECT_T jobj) {
	_json_main->publish_event(name, jobj);
}

