/*

Copyright (C) 2015-2016 Shane Carr

This file is part of Octave.

Octave is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 3 of the License, or (at your
option) any later version.

Octave is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Octave; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

*/

#ifndef octave_json_link_h
#define octave_json_link_h

#include <list>
#include <string>

#include "octave-link.h"
#include "json-util.h"
#include "oct-mutex.h"

// Circular reference
class json_main;

// Thread-safe queue
template<typename T> class json_queue {
public:
  json_queue();
  ~json_queue();

  void enqueue(const T& value);
  T dequeue();

private:
  std::queue<T> _queue;
  octave_mutex _mutex;
};

class octave_json_link : public octave_link
{

public:

  octave_json_link (json_main* __json_main);

  ~octave_json_link (void);

  std::string do_request_input (const std::string& prompt);

  bool do_confirm_shutdown (void);
  bool do_exit (int status);

  bool do_copy_image_to_clipboard (const std::string& file);

  bool do_edit_file (const std::string& file);
  bool do_prompt_new_edit_file (const std::string& file);

  int do_message_dialog (const std::string& dlg, const std::string& msg,
                         const std::string& title);

  std::string
  do_question_dialog (const std::string& msg, const std::string& title,
                      const std::string& btn1, const std::string& btn2,
                      const std::string& btn3, const std::string& btndef);

  std::pair<std::list<int>, int>
  do_list_dialog (const std::list<std::string>& list,
                  const std::string& mode,
                  int width, int height,
                  const std::list<int>& initial_value,
                  const std::string& name,
                  const std::list<std::string>& prompt,
                  const std::string& ok_string,
                  const std::string& cancel_string);

  std::list<std::string>
  do_input_dialog (const std::list<std::string>& prompt,
                   const std::string& title,
                   const std::list<float>& nr,
                   const std::list<float>& nc,
                   const std::list<std::string>& defaults);

  std::list<std::string>
  do_file_dialog (const filter_list& filter, const std::string& title,
                  const std::string &filename, const std::string &pathname,
                  const std::string& multimode);

  int
  do_debug_cd_or_addpath_error (const std::string& file,
                                const std::string& dir,
                                bool addpath_option);

  void do_change_directory (const std::string& dir);

  void do_execute_command_in_terminal (const std::string& command);

  void do_set_workspace (bool top_level, bool debug,
                         const std::list<workspace_element>& ws);

  void do_clear_workspace (void);

  void do_set_history (const string_vector& hist);
  void do_append_history (const std::string& hist_entry);
  void do_clear_history (void);

  void do_pre_input_event (void);
  void do_post_input_event (void);

  void do_enter_debugger_event (const std::string& file, int line);
  void do_execute_in_debugger_event (const std::string& file, int line);
  void do_exit_debugger_event (void);

  void do_update_breakpoint (bool insert, const std::string& file, int line);

  void do_set_default_prompts (std::string& ps1, std::string& ps2,
                               std::string& ps4);

  void do_show_preferences (void);

  void do_show_doc (const std::string& file);

  void do_show_static_plot (const std::string& term,
                            const std::string& content);

  // Custom methods
  void receive_message (const std::string& name, JSON_OBJECT_T jobj);

private:
  json_main* _json_main;
  void _publish_event (const std::string& name, JSON_OBJECT_T jobj);

  // Queues
  json_queue<std::string> request_input_queue;
  json_queue<bool> confirm_shutdown_queue;
  json_queue<bool> prompt_new_edit_file_queue;
  json_queue<int> message_dialog_queue;
  json_queue<std::string> question_dialog_queue;
  json_queue<std::pair<std::list<int>, int> > list_dialog_queue;
  json_queue<std::list<std::string> > input_dialog_queue;
  json_queue<std::list<std::string> > file_dialog_queue;
  json_queue<int> debug_cd_or_addpath_error_queue;
};

// Template classes require definitions in the header file...

template<typename T>
json_queue<T>::json_queue() { }

template<typename T>
json_queue<T>::~json_queue() { }

template<typename T>
void json_queue<T>::enqueue(const T& value) {
  _mutex.lock();
  _queue.push(value);
  _mutex.cond_signal();
  _mutex.unlock();
}

template<typename T>
T json_queue<T>::dequeue() {
  _mutex.lock();
  while (_queue.empty()) {
    _mutex.cond_wait();
  }
  T value = _queue.front();
  _queue.pop();
  _mutex.unlock();
  return value;
}

#endif
